@@ -11,19 +11,18 @@
     Implements support for high-level access to HDF5 groups.
 """
 
-from __future__ import absolute_import
-
 import posixpath as pp
-import six
 import numpy
 
+
 from .compat import filename_decode, filename_encode
 
-from .. import h5g, h5i, h5o, h5r, h5t, h5l, h5p
+from .. import h5, h5g, h5i, h5o, h5r, h5t, h5l, h5p, h5s, h5d
 from . import base
 from .base import HLObject, MutableMappingHDF5, phil, with_phil
 from . import dataset
 from . import datatype
+from .vds import vds_support
 
 
 class Group(HLObject, MutableMappingHDF5):
@@ -37,17 +36,33 @@ class Group(HLObject, MutableMappingHDF5):
         with phil:
             if not isinstance(bind, h5g.GroupID):
                 raise ValueError("%s is not a GroupID" % bind)
-            HLObject.__init__(self, bind)
+            super(Group, self).__init__(bind)
 
-    def create_group(self, name):
+
+    _gcpl_crt_order = h5p.create(h5p.GROUP_CREATE)
+    _gcpl_crt_order.set_link_creation_order(
+        h5p.CRT_ORDER_TRACKED | h5p.CRT_ORDER_INDEXED)
+    _gcpl_crt_order.set_attr_creation_order(
+        h5p.CRT_ORDER_TRACKED | h5p.CRT_ORDER_INDEXED)
+
+
+    def create_group(self, name, track_order=None):
         """ Create and return a new subgroup.
 
         Name may be absolute or relative.  Fails if the target name already
         exists.
+
+        track_order
+            Track dataset/group/attribute creation order under this group
+            if True. If None use global default h5.get_config().track_order.
         """
+        if track_order is None:
+            track_order = h5.get_config().track_order
+
         with phil:
             name, lcpl = self._e(name, lcpl=True)
-            gid = h5g.create(self.id, name, lcpl=lcpl)
+            gcpl = Group._gcpl_crt_order if track_order else None
+            gid = h5g.create(self.id, name, lcpl=lcpl, gcpl=gcpl)
             return Group(gid)
 
     def create_dataset(self, name, shape=None, dtype=None, data=None, **kwds):
@@ -70,10 +85,12 @@ class Group(HLObject, MutableMappingHDF5):
         Keyword-only arguments:
 
         chunks
-            (Tuple) Chunk shape, or True to enable auto-chunking.
+            (Tuple or int) Chunk shape, or True to enable auto-chunking. Integers can
+            be used for 1D shape.
+
         maxshape
-            (Tuple) Make the dataset resizable up to this shape.  Use None for
-            axes you want to be unlimited.
+            (Tuple or int) Make the dataset resizable up to this shape. Use None for
+            axes you want to be unlimited. Integers can be used for 1D shape.
         compression
             (String or int) Compression strategy.  Legal values are 'gzip',
             'szip', 'lzf'.  If an integer in range(10), this indicates gzip
@@ -101,12 +118,90 @@ class Group(HLObject, MutableMappingHDF5):
             (Scalar) Use this value for uninitialized parts of the dataset.
         track_times
             (T/F) Enable dataset creation timestamps.
+        track_order
+            (T/F) Track attribute creation order if True. If omitted use
+            global default h5.get_config().track_order.
+        external
+            (Iterable of tuples) Sets the external storage property, thus
+            designating that the dataset will be stored in one or more
+            non-HDF5 files external to the HDF5 file.  Adds each tuple
+            of (name, offset, size) to the dataset's list of external files.
+            Each name must be a str, bytes, or os.PathLike; each offset and
+            size, an integer.  If only a name is given instead of an iterable
+            of tuples, it is equivalent to [(name, 0, h5py.h5f.UNLIMITED)].
+        allow_unknown_filter
+            (T/F) Do not check that the requested filter is available for use.
+            This should only be used with ``write_direct_chunk``, where the caller
+            compresses the data before handing it to h5py.
         """
+        if 'track_order' not in kwds:
+            kwds['track_order'] = h5.get_config().track_order
+
         with phil:
-            dsid = dataset.make_new_dset(self, shape, dtype, data, **kwds)
+            group = self
+            if name:
+                if '/' in name:
+                    h5objects = [obj for obj in name.split('/') if len(obj)]
+                    name = h5objects[-1]
+                    h5objects = h5objects[:-1]
+
+                    for new_group in h5objects:
+                        group = group.get(new_group) or group.create_group(new_group)
+
+                name = self._e(name)
+
+            dsid = dataset.make_new_dset(group, shape, dtype, data, name, **kwds)
             dset = dataset.Dataset(dsid)
-            if name is not None:
-                self[name] = dset
+            return dset
+
+    if vds_support:
+        def create_virtual_dataset(self, name, layout, fillvalue=None):
+            """Create a new virtual dataset in this group.
+
+            See virtual datasets in the docs for more information.
+
+            name
+                (str) Name of the new dataset
+
+            layout
+                (VirtualLayout) Defines the sources for the virtual dataset
+
+            fillvalue
+                The value to use where there is no data.
+
+            """
+            from .vds import VDSmap
+            # Encode filenames and dataset names appropriately.
+            sources = []
+            for vspace, file_name, dset_name, src_space in layout.sources:
+                if file_name == self.file.filename:
+                    # use relative path if the source dataset is in the same
+                    # file, in order to keep the virtual dataset valid in case
+                    # the file is renamed.
+                    file_name = '.'
+                sources.append(VDSmap(vspace, filename_encode(file_name),
+                                      self._e(dset_name), src_space))
+
+            with phil:
+                group = self
+
+                if name:
+                    if '/' in name:
+                        h5objects = [obj for obj in name.split('/') if len(obj)]
+                        name = h5objects[-1]
+                        h5objects = h5objects[:-1]
+
+                        for new_group in h5objects:
+                            group = group.get(new_group) or group.create_group(new_group)
+
+                    name = self._e(name)
+
+                dsid = dataset.make_new_virtual_dset(group, layout.shape,
+                         sources=sources, dtype=layout.dtype, name=name,
+                         maxshape=layout.maxshape, fillvalue=fillvalue)
+
+                dset = dataset.Dataset(dsid)
+
             return dset
 
     def require_dataset(self, name, shape, dtype, exact=False, **kwds):
@@ -126,6 +221,9 @@ class Group(HLObject, MutableMappingHDF5):
             if not name in self:
                 return self.create_dataset(name, *(shape, dtype), **kwds)
 
+            if isinstance(shape, int):
+                shape = (shape,)
+
             dset = self[name]
             if not isinstance(dset, dataset.Dataset):
                 raise TypeError("Incompatible object (%s) already exists" % dset.__class__.__name__)
@@ -141,8 +239,41 @@ class Group(HLObject, MutableMappingHDF5):
 
             return dset
 
+    def create_dataset_like(self, name, other, **kwupdate):
+        """ Create a dataset similar to `other`.
+
+        name
+            Name of the dataset (absolute or relative).  Provide None to make
+            an anonymous dataset.
+        other
+            The dataset which the new dataset should mimic. All properties, such
+            as shape, dtype, chunking, ... will be taken from it, but no data
+            or attributes are being copied.
+
+        Any dataset keywords (see create_dataset) may be provided, including
+        shape and dtype, in which case the provided values take precedence over
+        those from `other`.
+        """
+        for k in ('shape', 'dtype', 'chunks', 'compression',
+                  'compression_opts', 'scaleoffset', 'shuffle', 'fletcher32',
+                  'fillvalue'):
+            kwupdate.setdefault(k, getattr(other, k))
+        # TODO: more elegant way to pass these (dcpl to create_dataset?)
+        dcpl = other.id.get_create_plist()
+        kwupdate.setdefault('track_times', dcpl.get_obj_track_times())
+        kwupdate.setdefault('track_order', dcpl.get_attr_creation_order() > 0)
+
+        # Special case: the maxshape property always exists, but if we pass it
+        # to create_dataset, the new dataset will automatically get chunked
+        # layout. So we copy it only if it is different from shape.
+        if other.maxshape != other.shape:
+            kwupdate.setdefault('maxshape', other.maxshape)
+
+        return self.create_dataset(name, **kwupdate)
+
     def require_group(self, name):
-        """ Return a group, creating it if it doesn't exist.
+        # TODO: support kwargs like require_dataset
+        """Return a group, creating it if it doesn't exist.
 
         TypeError is raised if something with that name already exists that
         isn't a group.
@@ -170,7 +301,7 @@ class Group(HLObject, MutableMappingHDF5):
         if otype == h5i.GROUP:
             return Group(oid)
         elif otype == h5i.DATASET:
-            return dataset.Dataset(oid)
+            return dataset.Dataset(oid, readonly=(self.file.mode == 'r'))
         elif otype == h5i.DATATYPE:
             return datatype.Datatype(oid)
         else:
@@ -198,7 +329,6 @@ class Group(HLObject, MutableMappingHDF5):
 
         >>> cls = group.get('foo', getclass=True)
         >>> if cls == SoftLink:
-        ...     print '"foo" is a soft link!'
         """
         # pylint: disable=arguments-differ
 
@@ -288,7 +418,7 @@ class Group(HLObject, MutableMappingHDF5):
                 htype.commit(self.id, name, lcpl=lcpl)
 
             else:
-                ds = self.create_dataset(None, data=obj, dtype=base.guess_dtype(obj))
+                ds = self.create_dataset(None, data=obj)
                 h5o.link(ds.id, self.id, name, lcpl=lcpl)
 
         if do_link:
@@ -348,11 +478,12 @@ class Group(HLObject, MutableMappingHDF5):
 
        Example:
 
-        >>> f = File('myfile.hdf5')
-        >>> f.listnames()
+        >>> f = File('myfile.hdf5', 'w')
+        >>> f.create_group("MyGroup")
+        >>> list(f.keys())
         ['MyGroup']
         >>> f.copy('MyGroup', 'MyCopy')
-        >>> f.listnames()
+        >>> list(f.keys())
         ['MyGroup', 'MyCopy']
 
         """
@@ -367,6 +498,8 @@ class Group(HLObject, MutableMappingHDF5):
             if isinstance(dest, Group):
                 if name is not None:
                     dest_path = name
+                elif source_path == '.':
+                    dest_path = pp.basename(h5i.get_name(source.id))
                 else:
                     # copy source into dest group: dest_name/source_name
                     dest_path = pp.basename(h5i.get_name(source[source_path].id))
@@ -422,7 +555,7 @@ class Group(HLObject, MutableMappingHDF5):
 
         Returning None continues iteration, returning anything else stops
         and immediately returns that value from the visit method.  No
-        particular order of iteration within groups is guranteed.
+        particular order of iteration within groups is guaranteed.
 
         Example:
 
@@ -448,7 +581,7 @@ class Group(HLObject, MutableMappingHDF5):
 
         Returning None continues iteration, returning anything else stops
         and immediately returns that value from the visit method.  No
-        particular order of iteration within groups is guranteed.
+        particular order of iteration within groups is guaranteed.
 
         Example:
 
@@ -474,12 +607,10 @@ class Group(HLObject, MutableMappingHDF5):
             r = u"<Closed HDF5 group>"
         else:
             namestr = (
-                u'"%s"' % self.name
+                '"%s"' % self.name
             ) if self.name is not None else u"(anonymous)"
-            r = u'<HDF5 group %s (%d members)>' % (namestr, len(self))
+            r = '<HDF5 group %s (%d members)>' % (namestr, len(self))
 
-        if six.PY2:
-            return r.encode('utf8')
         return r
 
 
@@ -532,7 +663,8 @@ class ExternalLink(object):
 
     def __init__(self, filename, path):
         self._filename = filename_decode(filename_encode(filename))
-        self._path = str(path)
+        self._path = path
 
     def __repr__(self):
-        return '<ExternalLink to "%s" in file "%s"' % (self.path, self.filename)
+        return '<ExternalLink to "%s" in file "%s"' % (self.path,
+                                                       self.filename)
